# -*- coding: utf8 -*-
from abc import ABCMeta, abstractmethod


class Job(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def wait(self):
        """

        :return:
        """

    @classmethod
    def wait_all(cls, *futures):
        results = []
        for future in futures:
            results.append(future.wait())

        return results



# -*- coding: utf8 -*-
import json
import os
import sys
import logging
import requests

is_python2 = sys.version_info[0] == 2


#  taken from python 3.3 source code
def which(cmd, mode=os.F_OK | os.X_OK, path=None):
    """Given a command, mode, and a PATH string, return the path which
    conforms to the given mode on the PATH, or None if there is no such
    file.
    `mode` defaults to os.F_OK | os.X_OK. `path` defaults to the result
    of os.environ.get("PATH"), or can be overridden with a custom search
    path.
    """
    # Check that a given file can be accessed with the correct mode.
    # Additionally check that `file` is not a directory, as on Windows
    # directories pass the os.access check.
    def _access_check(fn, current_mode):
        return (os.path.exists(fn) and os.access(fn, current_mode)
                and not os.path.isdir(fn))

    # If we're given a path with a directory part, look it up directly rather
    # than referring to PATH directories. This includes checking relative to the
    # current directory, e.g. ./script
    if os.path.dirname(cmd):
        if _access_check(cmd, mode):
            return cmd
        return None

    if path is None:
        path = os.environ.get("PATH", os.defpath)
    if not path:
        return None
    path = path.split(os.pathsep)

    if sys.platform == "win32":
        # The current directory takes precedence on Windows.
        if os.curdir not in path:
            path.insert(0, os.curdir)

        # PATHEXT is necessary to check on Windows.
        pathext = os.environ.get("PATHEXT", "").split(os.pathsep)
        # See if the given file matches any of the expected path extensions.
        # This will allow us to short circuit when given "python.exe".
        # If it does match, only test that one, otherwise we have to try
        # others.
        if any(cmd.lower().endswith(ext.lower()) for ext in pathext):
            files = [cmd]
        else:
            files = [cmd + ext for ext in pathext]
    else:
        # On other platforms you don't have things like PATHEXT to tell you
        # what file suffixes are executable, so just pass on cmd as-is.
        files = [cmd]

    seen = set()
    for current_dir in path:
        normal_dir = os.path.normcase(current_dir)
        if normal_dir not in seen:
            seen.add(normal_dir)
            for the_file in files:
                name = os.path.join(current_dir, the_file)
                if _access_check(name, mode):
                    return name
    return None


def pip_install(pip_server, require_package, user_path):
    from subprocess import Popen, PIPE

    pip_bin_path = which('pip')
    if pip_bin_path is None:
        python_bin_path = sys.executable
        pip_bin_path = os.path.join(os.path.dirname(python_bin_path), 'pip')
        if not os.path.exists(pip_bin_path):
            logging.warning("pip not found, can't self update missinglink sdk")
            return None, None

    if user_path:
        args = (pip_bin_path, 'install', '--upgrade', '--user', '-i', pip_server, require_package)
    else:
        args = (pip_bin_path, 'install', '--upgrade', '-i', pip_server, require_package)

    # noinspection PyBroadException
    return Popen(args, stdin=PIPE, stdout=PIPE, stderr=PIPE), args


def get_pip_server(keywords):
    pypi_server_hostname = 'testpypi' if 'test' in keywords else 'pypi'

    return 'https://{hostname}.python.org/pypi'.format(hostname=pypi_server_hostname)


def get_latest_pip_version(package, keywords, throw_exception=False):
    try:
        pypi_server = get_pip_server(keywords)

        url = '{server}/{package}/json'.format(server=pypi_server, package=package)
        r = requests.get(url)

        r.raise_for_status()

        package_info = json.loads(r.text)

        return package_info['info']['version']
    except Exception as e:
        if throw_exception:
            raise

        logging.exception('could not check for new missinglink-sdk version:\n%s', e)
        return None



class InconsistentQueryException(Exception):
    """Raised when a query have a problem in its structure
    """


class OrAndAndOnSameLevel(InconsistentQueryException):
    """
    Raised when a OR and a AND are on the same level as we don't know how to
    handle this case
    """


class NestedSearchFieldException(InconsistentQueryException):
    """
    Raised when a SearchField is nested in an other SearchField as it doesn't
    make sense. For Instance field1:(spam AND field2:eggs)
    """


class ObjectSearchFieldException(InconsistentQueryException):
    """
    Raised when a doted field name is queried which is not an object field
    """



# -*- coding: utf8 -*-

from click import option, Choice


def project_id_option(required=False):
    def decorator(f):
        return option('--projectId', '-p', type=int, metavar='<int>', required=required,
                      help='The project ID to find experiments. Use `mali projects list` to find your project IDs')(f)

    return decorator


def experiment_id_option(required=False):
    def decorator(f):
        return option('--experimentId', '-e', type=int, metavar='<int>', required=required,
                      help='The experiment ID. Use `mali experiments list` to find your experiment IDs')(f)

    return decorator


def chart_scope_option(required=False):
    def decorator(f):
        return option('--chartScope', '-cs', type=Choice(['test', 'validation', 'train']), required=required,
                      default='test', help='Scope type.')(f)

    return decorator


def chart_type_option(required=False):
    def decorator(f):
        return option('--chartType', '-ct', type=Choice(['line']), required=required,
                      default='line', help='Graph type.')(f)

    return decorator


def chart_name_option(required=False):
    def decorator(f):
        return option('--chartName', '-c', metavar='<str>', required=required,
                      help='The name of the chart. The name is used in order to identify the chart against different '
                           'experiments and through the same experiment.')(f)

    return decorator


def chart_x_option(required=False):
    def decorator(f):
        return option('--chartX', '-cx', metavar='<json_string>', required=required,
                      help='Array of m data points (X axis), Can be Strings, Integers or Floats.')(f)

    return decorator


def chart_y_option(required=False):
    def decorator(f):
        return option('--chartY', '-cy', metavar='<json_string>', required=required,
                      help='Array/Matrix of m data values. Can be either array m of Integers/Floats or a matrix (m arrays having n Ints/Floats each),  a full matrix describing the values of every chart in every data point')(f)

    return decorator


def chart_y_name_option(required=False):
    def decorator(f):
        return option('--chartYName', '-cyn', metavar='<json_str>', required=required, default='Y',
                      help='Display name for chart(s) Y axis')(f)

    return decorator


def chart_x_name_option(required=False):
    def decorator(f):
        return option('--chartXName', '-cxn', metavar='<str>', required=required, default='X',
                      help='Display name for charts X axis')(f)

    return decorator


def metrics_option(required=False):
    def decorator(f):
        return option('--metrics', '-m', metavar='<json_string>', required=required,
                      help='The metrics of the experiment as a jsonified string. The key should be the metric '
                           'name with "ex" prefix e.g. "ex_cost". The value is the metric value in String, Float, '
                           'Integer or Boolean.')(f)

    return decorator


def model_weights_hash_option(required=False):
    def decorator(f):
        return option('--weightsHash', '-wh', metavar='<sha1_hex>', required=required,
                      help="The hexadecimal sha1 hash of the model's weights")(f)

    return decorator



# -*- coding: utf8 -*-
import tempfile

import fudge
import six
from fudge import patched_context
from fudge.inspector import arg

from mali_commands.data_volume import create_data_volume, with_repo
from mali_commands.data_volume_config import DataVolumeConfig
from mali_commands.legit.dulwich import porcelain
from mali_commands.legit.object_store.gcs.gcs_object_store import GCSObjectStore, do_upload
from tests.base import BaseTest
from mali_commands.legit.dulwich.repo import Repo


class TestBGCSStorage(BaseTest):
    @classmethod
    def get_repo(cls, temp_path=None, volume_id=None, linked=False, description=None, display_name=None):
        temp_path = temp_path or tempfile.tempdir
        volume_id = volume_id or cls.some_random_shit_number_int63()
        data_path = temp_path
        description = description or cls.some_random_shit(size=50)
        display_name = display_name or cls.some_random_shit(size=10)
        res = create_data_volume(volume_id, data_path, linked, display_name, description)
        return res

    @classmethod
    def __base_64_decoder_func(cls):
        import base64
        if six.PY2:
            return base64.decodestring
        else:
            return base64.decodebytes

    @classmethod
    def image(cls):
        return cls.__base_64_decoder_func()(
            b'/9j/4AAQSkZJRgABAQEAYABgAAD//gA+Q1JFQVRPUjogZ2QtanBlZyB2MS4wICh1c2luZyBJSkcgSlBFRyB2ODApL'
            b'CBkZWZhdWx0IHF1YWxpdHkK/9sAQwAIBgYHBgUIBwcHCQkICgwUDQwLCwwZEhMPFB0aHx4dGhwcICQuJyAiLCMcHCg'
            b'3KSwwMTQ0NB8nOT04MjwuMzQy/9sAQwEJCQkMCwwYDQ0YMiEcITIyMjIyMjIyMjIyMjIyMjIyMjIyMjIyMjIyMjIyMj'
            b'IyMjIyMjIyMjIyMjIyMjIyMjIy/8AAEQgAHgAeAwEiAAIRAQMRAf/EAB8AAAEFAQEBAQEBAAAAAAAAAAABAgMEBQYHC'
            b'AkKC//EALUQAAIBAwMCBAMFBQQEAAABfQECAwAEEQUSITFBBhNRYQcicRQygZGhCCNCscEVUtHwJDNicoIJChYXGBka'
            b'JSYnKCkqNDU2Nzg5OkNERUZHSElKU1RVVldYWVpjZGVmZ2hpanN0dXZ3eHl6g4SFhoeIiYqSk5SVlpeYmZqio6Slpqe'
            b'oqaqys7S1tre4ubrCw8TFxsfIycrS09TV1tfY2drh4uPk5ebn6Onq8fLz9PX29/j5+v/EAB8BAAMBAQEBAQEBAQEAAAA'
            b'AAAABAgMEBQYHCAkKC//EALURAAIBAgQEAwQHBQQEAAECdwABAgMRBAUhMQYSQVEHYXETIjKBCBRCkaGxwQkjM1LwFWJy'
            b'0QoWJDThJfEXGBkaJicoKSo1Njc4OTpDREVGR0hJSlNUVVZXWFlaY2RlZmdoaWpzdHV2d3h5eoKDhIWGh4iJipKTlJWWl'
            b'5iZmqKjpKWmp6ipqrKztLW2t7i5usLDxMXGx8jJytLT1NXW19jZ2uLj5OXm5+jp6vLz9PX29/j5+v/aAAwDAQACEQMRAD8'
            b'ArUVLFbSTIWjCnbyRuAwPX6U14XR9pXJ46c5yMiv0PmV7H5ryu1xlFSxW0kyFowp28kbgMD1+lMkjaJ9rYzgHg54IyKLq9g5'
            b'WlclhungRkUKVYYYHPI/OgXUiuzKFG5AhGP4cYx+gqCijlQc8tieG6eBGRQpVhhgc8j86jkkMr7iAOAMD0AwP5UyihRSdwcm1Y//Z')

    @classmethod
    def make_temp_files(cls, count, image=False):
        suffix = '.txt'
        if image:
            suffix = '.jpg'
            # image_bytes = cls.__base_64_decoder_func(cls.image_str)
        for i in range(count):
            f = tempfile.NamedTemporaryFile(suffix=suffix, prefix='test_enumerate_paths__', delete=False)
            if image:
                f.write(cls.image())
            else:
                f.write(bytearray("Hello World!\n{}\n".format(i), 'utf8'))
            f.close()
            yield f.name

    def test_group_files_by_meta_jpg(self):
        file_count = 10
        is_embedded = True
        volume_id = self.some_random_shit_number_int63()
        repo_config = self.get_repo(volume_id=volume_id, linked=is_embedded)
        tempfiles = list(self.make_temp_files(file_count, image=True))
        with with_repo(None, volume_id) as rep:
            rep.data_path = '/'
            rep.path = '/'

            closure_dict = {}

            def patched_stage(inst, paths, embedded, callback):
                for path in paths:
                    self.assertIn('/' + path, tempfiles)
                self.assertEqual(len(paths), len(tempfiles))
                closure_dict['rel_paths'] = paths

            with patched_context(rep, 'stage', patched_stage):
                porcelain.add(rep, tempfiles, is_embedded)

            relpaths = closure_dict['rel_paths']
            index = rep.open_index()
            files = list(rep._build_stage_files(relpaths, index, is_embedded))

            object_store = rep.create_object_store()
            self.assertIsNotNone(object_store)
            # object_store.add_objects(files)
            res = object_store._group_files_by_meta(files)
            res = object_store._group_files_by_meta(files)
            self.assertIsNotNone(res)
            keys = list(res.keys())
            self.assertTrue(len(keys) == 1)
            self.assertEquals(keys[0], 'image/jpeg')
            self.assertEquals(len(res[keys[0]]), file_count)

    def test_group_files_by_meta_txt(self):
        file_count = 10
        is_embedded = True
        volume_id = self.some_random_shit_number_int63()
        repo_config = self.get_repo(volume_id=volume_id, linked=is_embedded)
        tempfiles = list(self.make_temp_files(file_count, image=False))
        with with_repo(None, volume_id) as rep:
            rep.data_path = '/'
            rep.path = '/'

            closure_dict = {}

            def patched_stage(inst, paths, embedded, callback):
                for path in paths:
                    self.assertIn('/' + path, tempfiles)
                self.assertEqual(len(paths), len(tempfiles))
                closure_dict['rel_paths'] = paths

            with patched_context(rep, 'stage', patched_stage):
                porcelain.add(rep, tempfiles, is_embedded)

            relpaths = closure_dict['rel_paths']
            index = rep.open_index()
            files = list(rep._build_stage_files(relpaths, index, is_embedded))

            object_store = rep.create_object_store()
            self.assertIsNotNone(object_store)
            # object_store.add_objects(files)
            res = object_store._group_files_by_meta(files)
            self.assertIsNotNone(res)
            keys = list(res.keys())
            self.assertTrue(len(keys) == 1)
            self.assertIsNone(keys[0])

            self.assertEquals(len(res[keys[0]]), file_count)

    def test_group_files_by_meta_mixed(self):
        per_meta_file_count = 10
        is_embedded = True
        volume_id = self.some_random_shit_number_int63()
        repo_config = self.get_repo(volume_id=volume_id, linked=is_embedded)
        temp_txt = list(self.make_temp_files(per_meta_file_count, image=False))
        temp_images = list(self.make_temp_files(per_meta_file_count, image=True))
        tempfiles = temp_txt + temp_images
        with with_repo(None, volume_id) as rep:
            rep.data_path = '/'
            rep.path = '/'

            closure_dict = {}

            def patched_stage(inst, paths, embedded, callback):
                for path in paths:
                    self.assertIn('/' + path, tempfiles)
                self.assertEqual(len(paths), len(tempfiles))
                closure_dict['rel_paths'] = paths

            with patched_context(rep, 'stage', patched_stage):
                porcelain.add(rep, tempfiles, is_embedded)

            relpaths = closure_dict['rel_paths']
            index = rep.open_index()
            files = list(rep._build_stage_files(relpaths, index, is_embedded))

            object_store = rep.create_object_store()
            self.assertIsNotNone(object_store)
            # object_store.add_objects(files)
            res = object_store._group_files_by_meta(files)
            self.assertIsNotNone(res)
            keys = list(res.keys())
            self.assertTrue(len(keys) == 2)
            self.assertIn(None, res)
            self.assertIn('image/jpeg', res)

            self.assertEquals(len(res[keys[0]]), per_meta_file_count)
            self.assertEquals(len(res[keys[1]]), per_meta_file_count)

    @fudge.patch('mali_commands.legit.object_store.gcs.backend_gcs_object_store.BackendGCSSignedUrlService.get_signed_urls')
    def test_get_urls_for_paths(self, get_signed_urls):
        file_count = 10
        is_embedded = True
        volume_id = self.some_random_shit_number_int63()
        repo_config = self.get_repo(volume_id=volume_id, linked=is_embedded)
        tempfiles = list(self.make_temp_files(file_count, image=True))
        with with_repo(None, volume_id) as rep:
            rep.data_path = '/'
            rep.path = '/'

            closure_dict = {}

            def patched_stage(inst, paths, embedded, callback):
                for path in paths:
                    self.assertIn('/' + path, tempfiles)
                self.assertEqual(len(paths), len(tempfiles))
                closure_dict['rel_paths'] = paths

            with patched_context(rep, 'stage', patched_stage):
                porcelain.add(rep, tempfiles, is_embedded)

            relpaths = closure_dict['rel_paths']
            index = rep.open_index()
            files = list(rep._build_stage_files(relpaths, index, is_embedded))

            object_store = rep.create_object_store()
            self.assertIsNotNone(object_store)
            # object_store.add_objects(files)
            res = object_store._group_files_by_meta(files)

            content_type = 'image/jpeg'
            sign_files = res[content_type]
            content_headers = object_store.get_content_headers()
            get_signed_urls.expects_call().with_args(['HEAD', 'PUT'], sign_files, content_type, **content_headers).returns({'HEAD': sign_files, 'PUT': sign_files}).times_called(1)
            # get_signed_urls.expects_call().returns((access_token, refresh_token, id_token)).times_called(1)
            head, put = object_store._get_urls_for_paths(sign_files, content_type, content_headers)
            self.assertIs(head, sign_files)
            self.assertIs(put, sign_files)

    def test_call_for_upload(self):
        file_count = 10
        is_embedded = True
        volume_id = self.some_random_shit_number_int63()
        repo_config = self.get_repo(volume_id=volume_id, linked=is_embedded)
        tempfiles = list(self.make_temp_files(file_count, image=True))
        with with_repo(None, volume_id) as rep:
            rep.data_path = '/'
            rep.path = '/'

            closure_dict = {}

            def patched_stage(inst, paths, embedded, callback):
                for path in paths:
                    self.assertIn('/' + path, tempfiles)
                self.assertEqual(len(paths), len(tempfiles))
                closure_dict['rel_paths'] = paths

            with patched_context(rep, 'stage', patched_stage):
                porcelain.add(rep, tempfiles, is_embedded)

            relpaths = closure_dict['rel_paths']
            index = rep.open_index()
            files = list(rep._build_stage_files(relpaths, index, is_embedded))

            object_store = rep.create_object_store()
            self.assertIsNotNone(object_store)
            res = object_store._group_files_by_meta(files)

            content_type = 'image/jpeg'
            sign_files = list(res[content_type])
            content_headers_full = object_store.get_content_headers(content_type)

            def validate_upload_command(inst, obj, content_type_arg, head_url, put_url):
                self.assertTrue(obj in sign_files)

                x = GCSObjectStore._get_shafile_path(obj.blob.id)
                self.assertEquals(content_type_arg, content_type)
                self.assertEquals('{}_HEAD'.format(x), head_url)
                self.assertEquals('{}_PUT'.format(x), put_url)

                # self.upload(cur_file.blob, content_type, head_url, put_url, self.get_content_headers(content_type))

                pass

            def build_upload_paths(inst, paths, content_type, headers):
                head_urls = list(map(lambda x: '{}_HEAD'.format(x), paths))
                put_urls = list(map(lambda x: '{}_PUT'.format(x), paths))
                return head_urls, put_urls

            callback = None

            with patched_context(object_store, 'upload', validate_upload_command):
                with patched_context(object_store, '_get_urls_for_paths', build_upload_paths):
                    object_store.add_objects(files, callback)

    @fudge.patch('mali_commands.legit.object_store.gcs.gcs_object_store.do_upload')
    def test_call_to_upload_file(self, do_upload_patched):
        volume_id = self.some_random_shit_number_int63()
        temp_files = list(self.make_temp_files(1, image=True))[0]
        upload_file_request = Repo.UploadFileRequest(temp_files, temp_files[1:], True)

        headers = arg.isinstance(dict)

        file_upload_path = str(volume_id) + '/' + GCSObjectStore._get_shafile_path(upload_file_request.blob.id)
        head_url = 'HEAD'
        put_url = 'PUT'
        auth_arg = None
        bucket_name_arg = None
        credentials = arg.any()
        do_upload_patched.expects_call().with_args(
            auth_arg,
            bucket_name_arg,
            file_upload_path,
            upload_file_request.full_path,
            headers,
            head_url,
            put_url,
            credentials,
        ).returns(None).times_called(1)

        fake_object_store_config = {}
        fake_data_volume_config = fudge.Fake().has_attr(object_store_config=fake_object_store_config, volume_id=volume_id)
        fake_connection = fudge.Fake().has_attr(data_volume_config=fake_data_volume_config)
        object_store = GCSObjectStore(fake_connection, use_multiprocess=False)

        content_type = 'image/jpeg'

        object_store.upload(upload_file_request, content_type, head_url, put_url)

    @fudge.patch('mali_commands.legit.gcs_utils.GCSUpload.upload')
    def test_do_upload(self, upload):

        volume_id = self.some_random_shit_number_int63()
        temp_files = list(self.make_temp_files(1, image=True))[0]
        upload_file_request = Repo.UploadFileRequest(temp_files, temp_files[1:], True)

        content_type = 'image/jpeg'

        file_upload_path = str(volume_id) + '/' + GCSObjectStore._get_shafile_path(upload_file_request.blob.id)
        head_url = 'head_url_' + self.some_random_shit()
        put_url = 'put_url_' + self.some_random_shit()

        auth = 'auth_' + self.some_random_shit()
        bucket_name = 'bucket_name_' + self.some_random_shit()
        headers = {}
        callback = None

        upload.expects_call().with_args(
            arg.any(),
            headers,
            callback
        ).returns(None).times_called(1)

        do_upload(auth, bucket_name, file_upload_path, upload_file_request.full_path, headers, head_url, put_url)

    @fudge.patch('mali_commands.legit.gcs_utils.GCSUploadDirect.upload')
    def test_do_upload_directUpload(self, upload_direct):

        volume_id = self.some_random_shit_number_int63()
        temp_files = list(self.make_temp_files(1, image=True))[0]
        upload_file_request = Repo.UploadFileRequest(temp_files, temp_files[1:], True)

        file_upload_path = str(volume_id) + '/' + GCSObjectStore._get_shafile_path(upload_file_request.blob.id)
        head_url = None
        put_url = None

        auth = 'auth_' + self.some_random_shit()
        bucket_name = 'bucket_name_' + self.some_random_shit()

        headers = {}

        upload_direct.expects_call().with_args(
            bucket_name, file_upload_path, arg.any(), headers
        ).returns(None).times_called(1)

        do_upload(auth, bucket_name, file_upload_path, upload_file_request.full_path, headers, head_url, put_url)



# -*- coding: utf8 -*-
from .base_connection import BaseConnection


class DatastoreSqlHelper(object):
    def __init__(self):
        pass


class DatastoreConnection(BaseConnection):
    def __init__(self, data_volume_config, project, **kwargs):
        self.__project = project
        super(DatastoreConnection, self).__init__(data_volume_config, **kwargs)

    def _create_connection(self, **kwargs):
        from google.cloud import datastore

        return datastore.Client(project=self.__project, _use_grpc=False)

    @classmethod
    def create_entity(cls, key, exclude_from_indexes=None):
        from google.cloud import datastore
        return datastore.Entity(key, exclude_from_indexes=exclude_from_indexes or ())

    def _create_cursor(self):
        return self._native_conn

    def _commit(self):
        pass

    def _rollback(self):
        pass

    def create_sql_helper(self):
        return DatastoreSqlHelper()



# -*- coding: utf8 -*-
import os

import shutil

from mali_commands.data_volume import update_config_file
from tests_data_management.base import BaseTest
from mali_commands.utilities.os_utils import flatten_dir, remove_dir


class TestDataManagement(BaseTest):
    def testUpload(self):
        self.populate_data([
            'foo.txt',
        ])

        volume_id = self.create_data_volume()
        self.add_file(volume_id, ['foo.txt'])
        self.add_file(volume_id, ['foo.txt'])

    def testUploadSingleProcess(self):
        self.populate_data([
            'foo.txt',
        ])

        params = {
            'object_store': {
                'use_multiprocess': False,
            }
        }

        volume_id = self.create_data_volume()

        update_config_file(volume_id, params)

        self.add_file(volume_id, ['foo.txt'])
        self.add_file(volume_id, ['foo.txt'])

    # def testUploadDirectBucket(self):
    #     self.populate_data([
    #         'foo.txt',
    #     ])
    #
    #     volume_id = self.create_data_volume()
    #     params = {
    #         'object_store': {
    #             'bucket_name': 'staging_data_volumes',
    #         }
    #     }
    #
    #     update_config_file(volume_id, params)
    #
    #     self.add_file(volume_id, ['foo.txt'])
    #     self.add_file(volume_id, ['foo.txt'])

    def testBasicFlow(self):
        """Test basic flow
        - Create a data volume
        - Add a file
        - Add a metadata
        - Commit
        - Clone
        """

        self.populate_data([
            'foo.txt',
        ])

        volume_id = self.create_data_volume()
        self.add_metadata(volume_id, ['foo.txt'], {'name': 'foo'})
        self.add_file(volume_id, ['foo.txt'])
        result = self.commit(volume_id)

        self.assertEqual(result['version_total_data_points'], '1')
        self.assertEqual(result['version_total_meta_data'], '1')
        self.assertEqual(result['total_meta_data'], '1')
        self.assertEqual(result['total_data_points'], '1')

        commit_id = result['commit_id']
        clone_path = os.path.join(self.clone_dir, '$@phase/$@name')
        self.clone(volume_id, clone_path, 'name:foo @version:%s' % commit_id)
        flatten_dir(self.clone_dir)

        self.assertEqualDir(self.data_dir, self.clone_dir)

    def testMultipleCommits(self):
        self.populate_data([
            'foo.txt',
            'bar.txt'
        ])

        volume_id = self.create_data_volume()
        self.add_file(volume_id, ['foo.txt'])
        self.add_metadata(volume_id, ['foo.txt'], {'name': 'foo'})
        result = self.commit(volume_id)

        self.assertEqual(result['version_total_data_points'], '1')
        self.assertEqual(result['version_total_meta_data'], '1')
        self.assertEqual(result['total_meta_data'], '1')
        self.assertEqual(result['total_data_points'], '1')

        self.add_file(volume_id, ['bar.txt'])
        self.add_metadata(volume_id, ['bar.txt'], {'name': 'foo'})
        result = self.commit(volume_id)

        self.assertEqual(result['version_total_data_points'], '1')
        self.assertEqual(result['version_total_meta_data'], '1')
        self.assertEqual(result['total_meta_data'], '2')
        self.assertEqual(result['total_data_points'], '2')

        commit_id = result['commit_id']

        clone_path = os.path.join(self.clone_dir, '$@phase/$@name')
        self.clone(volume_id, clone_path, '@version:%s' % commit_id)
        flatten_dir(self.clone_dir)

        self.assertEqualDir(self.data_dir, self.clone_dir)

    def testMultipleCommitsNoMetadata(self):
        self.populate_data([
            'foo.txt',
            'bar.txt'
        ])

        volume_id = self.create_data_volume()
        self.add_file(volume_id, ['foo.txt'])
        result = self.commit(volume_id)

        self.assertEqual(result['version_total_data_points'], '1')
        self.assertEqual(result['version_total_meta_data'], '0')
        self.assertEqual(result['total_meta_data'], '0')
        self.assertEqual(result['total_data_points'], '1')

        self.add_file(volume_id, ['bar.txt'])
        result = self.commit(volume_id)

        self.assertEqual(result['version_total_data_points'], '1')
        self.assertEqual(result['version_total_meta_data'], '0')
        self.assertEqual(result['total_meta_data'], '0')
        self.assertEqual(result['total_data_points'], '2')

    def testNoMetadata(self):
        self.populate_data(['foo.txt'])

        volume_id = self.create_data_volume()

        self.add_file(volume_id, ['foo.txt'])

        query_result_staging = self.query(volume_id, '@version:staging')

        self.assertIsNotNone(query_result_staging[0]['@id'])
        self.assertEqual(query_result_staging[0]['@path'], 'foo.txt')
        self.assertEqual(query_result_staging[0]['@size'], '4096')
        self.assertIsNotNone(query_result_staging[0]['@url'])
        self.assertEqual(query_result_staging[0]['@version'], 'staging')

        commit_result1 = self.commit(volume_id)

        query_result_after_commit = self.query(volume_id, '@version:%s' % commit_result1['commit_id'])

        self.assertIsNotNone(query_result_after_commit[0]['@id'])
        self.assertEqual(query_result_after_commit[0]['@path'], 'foo.txt')
        self.assertEqual(query_result_after_commit[0]['@size'], '4096')
        self.assertIsNotNone(query_result_after_commit[0]['@url'])
        self.assertEqual(query_result_after_commit[0]['@version'], commit_result1['commit_id'])

    def testModifyMeta(self):
        self.populate_data(['foo.txt'])

        volume_id = self.create_data_volume()
        self.add_file(volume_id, ['foo.txt'])
        self.add_metadata(volume_id, ['foo.txt'], {'state': 'ca'})
        commit_result1 = self.commit(volume_id)

        query_result1 = self.query(volume_id, '@version:%s state:ca' % commit_result1['commit_id'])

        self.assertEqual(1, len(query_result1))
        self.assertEqual('ca', query_result1[0]['state'])

        self.add_metadata(volume_id, ['foo.txt'], {'state': 'il'})
        commit_result2 = self.commit(volume_id)

        query_result2 = self.query(volume_id, '@version:%s state:il' % commit_result2['commit_id'])
        self.assertEqual(1, len(query_result2))
        self.assertEqual('il', query_result2[0]['state'])

        query_result2 = self.query(volume_id, '@version:%s state:ca' % commit_result2['commit_id'])
        self.assertEqual(0, len(query_result2))

    def testDuplicateMetaQueryStaging(self):
        self.populate_data(['foo.txt', 'bar.txt'])

        volume_id = self.create_data_volume()
        self.add_file(volume_id, ['foo.txt', 'bar.txt'])
        self.add_metadata(volume_id, ['foo.txt'], {'state': 'ca'})
        self.add_metadata(volume_id, ['foo.txt'], {'state': 'ca'})
        self.add_metadata(volume_id, ['bar.txt'], {'state': 'ca'})
        query_result1 = self.query(volume_id, '@version:staging state:ca')
        self.assertEqual(2, len(query_result1))
        self.assertEqual('ca', query_result1[0]['state'])
        self.assertEqual('ca', query_result1[1]['state'])

        self.assertEqual(sorted([item['@path'] for item in query_result1]), ['bar.txt', 'foo.txt'])

    def testClone_metadataFromAPreviousCommit(self):
        self.populate_data(['foo.txt', 'bar.txt'])

        volume_id = self.create_data_volume()

        self.add_file(volume_id, ['foo.txt'])
        self.add_metadata(volume_id, ['foo.txt'], {'name': 'foo'})
        commit_result1 = self.commit(volume_id)
        commit_id1 = commit_result1['commit_id']

        self.add_file(volume_id, ['bar.txt'])
        self.add_metadata(volume_id, ['bar.txt'], {'name': 'bar'})
        result = self.commit(volume_id)
        commit_id2 = result['commit_id']

        query_result = self.query(volume_id, '@version:%s name:foo' % commit_id1)
        self.assertEqual(len(query_result), 1)
        self.assertEqual(query_result[0]['@path'], 'foo.txt')

        query_result = self.query(volume_id, '@version:%s name:foo' % commit_id2)
        self.assertEqual(len(query_result), 1)
        self.assertEqual(query_result[0]['@path'], 'foo.txt')

    def testGroup(self):
        files = [
            '1.txt', '2.txt', '3.txt', '4.txt',
        ]
        self.populate_data(files)

        volume_id = self.create_data_volume()
        self.add_file(volume_id, files)
        self.add_metadata(volume_id, files, {'name': 'foo'})
        result = self.commit(volume_id)

        commit_id = result['commit_id']
        self.query(volume_id, 'name:foo @group:name @version:%s' % commit_id)

    def testClonedFilesWithNewMetadata(self):
        files = [
            '1.txt', '2.txt',
        ]

        self.populate_data(files)

        volume_id = self.create_data_volume()
        self.add_file(volume_id, files)
        self.add_metadata(volume_id, files, {'first': '1'})
        first_staging_query_result = self.query(volume_id, '@version:staging')

        self.assertEqual(2, len(first_staging_query_result))
        self.assertIsNotNone(first_staging_query_result[0]['@id'])
        self.assertIsNotNone(first_staging_query_result[1]['@id'])

        self.assertEqual(first_staging_query_result[0]['@path'], '1.txt')
        self.assertEqual(first_staging_query_result[1]['@path'], '2.txt')

        self.assertIsNotNone(first_staging_query_result[0]['@url'])
        self.assertIsNotNone(first_staging_query_result[1]['@url'])

        self.assertEqual(first_staging_query_result[0]['first'], '1')
        self.assertEqual(first_staging_query_result[1]['first'], '1')

        self.assertEqual(first_staging_query_result[0]['@version'], 'staging')
        self.assertEqual(first_staging_query_result[1]['@version'], 'staging')

        first_commit_result = self.commit(volume_id)

        self.assertEqual(first_commit_result['version_total_data_points'], '2')
        self.assertEqual(first_commit_result['version_total_meta_data'], '2')
        self.assertEqual(first_commit_result['total_meta_data'], '2')
        self.assertEqual(first_commit_result['total_data_points'], '2')

        first_version_query_result = self.query(volume_id, '@version:%s' % first_commit_result['commit_id'])

        self.assertEqual(2, len(first_version_query_result))

        cloned_files = []

        for filename in files:
            cloned_file = 'dup_' + filename

            cloned_files.append(cloned_file)
            shutil.copy(os.path.join(self.data_dir, filename), os.path.join(self.data_dir, cloned_file))

        self.add_file(volume_id, cloned_files)
        self.add_metadata(volume_id, cloned_files, {'second': '2'})

        second_staging_query_result = self.query(volume_id, '@version:staging')

        self.assertEqual(2, len(second_staging_query_result))
        self.assertIsNotNone(second_staging_query_result[0]['@id'])
        self.assertIsNotNone(second_staging_query_result[1]['@id'])

        self.assertEqual(second_staging_query_result[0]['@path'], 'dup_1.txt')
        self.assertEqual(second_staging_query_result[1]['@path'], 'dup_2.txt')

        self.assertIsNotNone(second_staging_query_result[0]['@url'])
        self.assertIsNotNone(second_staging_query_result[1]['@url'])

        self.assertEqual(second_staging_query_result[0]['first'], None)
        self.assertEqual(second_staging_query_result[1]['first'], None)
        self.assertEqual(second_staging_query_result[0]['second'], '2')
        self.assertEqual(second_staging_query_result[1]['second'], '2')

        self.assertEqual(second_staging_query_result[0]['@version'], 'staging')
        self.assertEqual(second_staging_query_result[1]['@version'], 'staging')

        second_commit_result = self.commit(volume_id)

        self.assertEqual(second_commit_result['version_total_data_points'], '2')
        self.assertEqual(second_commit_result['version_total_meta_data'], '2')
        self.assertEqual(second_commit_result['total_meta_data'], '4')
        self.assertEqual(second_commit_result['total_data_points'], '4')

        second_version_query_result = self.query(volume_id, '@version:%s' % second_commit_result['commit_id'])

        self.assertEqual(4, len(second_version_query_result))

    def testClone_changeFileContent_getLatestContent(self):
        volume_id = self.create_data_volume()
        filename = 'foo.txt'
        first_content = self.some_random_shit()
        print('First string:', first_content)
        second_content = self.some_random_shit()
        print('Second string:', second_content)
        clone_path = os.path.join(self.clone_dir, '$@phase/$@name')

        # First commit
        self.populate_data([filename], [first_content])
        self.add_file(volume_id, [filename])
        self.add_metadata(volume_id, [filename], {'name': filename})
        result = self.commit(volume_id)

        self.clone(volume_id, clone_path, '@version:%s' % result['commit_id'])
        flatten_dir(self.clone_dir)
        self.assertEqualDir(self.data_dir, self.clone_dir)
        remove_dir(self.clone_dir)

        # Second commit, change the file content
        self.populate_data([filename], [second_content])
        self.add_file(volume_id, [filename])
        result = self.commit(volume_id)

        self.clone(volume_id, clone_path, '@version:%s' % result['commit_id'])
        flatten_dir(self.clone_dir)
        self.assertEqualDir(self.data_dir, self.clone_dir)
        remove_dir(self.clone_dir)

        # Third commit, change the file content back to the original
        self.populate_data([filename], [first_content])
        self.add_file(volume_id, [filename])
        result = self.commit(volume_id)

        self.clone(volume_id, clone_path, '@version:%s' % result['commit_id'])
        flatten_dir(self.clone_dir)
        self.assertEqualDir(self.data_dir, self.clone_dir)
        remove_dir(self.clone_dir)



# -*- coding: utf8 -*-
import os
import errno
import six

from contextlib import closing
import click
from mali_commands.legit.dulwich.errors import NotGitRepository

from mali_commands.data_volume_config import DataVolumeConfig
from .repo import MlRepo
from mali_commands.config import Config


def default_data_volume_path(volume_id):
    return os.path.join(os.path.expanduser('~'), '.MissingLinkAI', 'data_volumes', str(volume_id))


def update_config_file(volume_id, params):
    path = default_data_volume_path(volume_id)

    try:
        os.makedirs(path)
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise

    config = DataVolumeConfig(path)

    config.update_and_save(params)

    return config


def create_data_volume(volume_id, data_path, linked, display_name, description, **kwargs):
    params = {
        'general': {
            'id': volume_id,
            'embedded': not linked,
            'datapath': data_path,
            'display_name': display_name,
            'description': description
        }
    }

    params.update(**kwargs)
    return update_config_file(volume_id, params)


def with_repo(config, volume_id, read_only=False):
    repo_root = default_data_volume_path(volume_id)

    try:
        return closing(MlRepo(config, repo_root, read_only=read_only))
    except NotGitRepository:
        msg = 'Data volume {0} was not mapped on this machine, you should call "mali data map {0} --dataPath [root path of data]" in order to work with the volume locally'.format(volume_id)
        click.echo(msg, err=True)
        exit(1)


def validate_repo_data_path(repo, volume_id):
    if not repo.data_path:
        msg = "Data Volume {volume_id} is not mapped to a local folder,\n" \
              "the data root is needed in order to add data to the data volume.\n" \
              "You can run:\n" \
              "mali data get {volume_id} --dataPath <data root folder>".format(volume_id=volume_id)
        click.echo(msg.strip(), err=True)
        exit(1)



# -*- coding: utf8 -*-
import glob
import json
import os
from uuid import uuid4
from contextlib import closing

import click
import errno

import datetime
import requests
import six
from botocore.exceptions import NoCredentialsError

from mali_commands.legit import MetadataOperationError
from tqdm import tqdm

from mali_commands.config import Config
from mali_commands.legit.dulwich import porcelain
from mali_commands.commons import add_to_data_if_not_none, output_result
from mali_commands.data_volume import create_data_volume, with_repo, validate_repo_data_path, default_data_volume_path
import shutil

from mali_commands.legit.gcs_utils import do_upload
from mali_commands.legit.multi_process_control import get_multi_process_control


@click.group('data')
def data_commands():
    pass


ignore_files = ['.DS_Store']


def enumerate_path(path):
    path = __expend_and_validate_path(path)
    if is_glob(path):
        for path in glob.glob(path):
            yield path

        return

    if os.path.isfile(path):
        yield path
        return

    if os.path.isdir(path):
        for root, dirs, files in os.walk(path, followlinks=True):
            for file_ in files:
                if file_ in ignore_files:
                    continue
                yield os.path.join(root, file_)
                # TODOx: ASK SHAY
                # elif path not in ignore_files:
                #     total_files += 1


def enumerate_paths(paths):
    for path in paths:
        for file_ in enumerate_path(path):
            yield file_


def get_total_files_in_path(paths):
    total_files = 0
    for _ in enumerate_paths(paths):
        total_files += 1

    return total_files


def is_glob(path):
    return '*' in path or '?' in path


def __expend_and_validate_path(path, expand_vars=True, validate_path=True, abs_path=True):
    path = os.path.expanduser(path)
    if expand_vars:
        path = os.path.expandvars(path)

    if abs_path:
        path = os.path.abspath(path)

    if not is_glob(path) and validate_path:
        if not os.path.exists(path):
            click.echo('not found %s' % path, err=True)
            exit(1)
            return

    return path


def get_batch_of_files_from_paths(paths, batch_size):
    batch = []
    for file_or_path in enumerate_paths(paths):
        batch.append(file_or_path)
        if len(batch) == batch_size:
            yield batch
            batch = []

    if len(batch) > 0:
        yield batch


@data_commands.command('map')
@click.argument('volumeId', envvar='VOLUMEID', type=int)
@click.option('--dataPath', required=True)
@click.pass_context
def _cmd_add_data_path(ctx, volumeid, datapath):
    result = ctx.obj.handle_api(ctx.obj, requests.get, "data_volumes/{volume_id}".format(volume_id=volumeid))

    datapath = __expend_and_validate_path(datapath)

    bucket_name = result.get('bucket_name')

    params = {}
    if bucket_name is not None:
        params['object_store'] = {'bucket_name': bucket_name}

    create_data_volume(
        result['id'],
        datapath,
        not result['embedded'],
        result.get('display_name'),
        result.get('description'),
        **params)

    display_name = result.get('display_name', 'No display name provided')
    click.echo('Initialized data volume %s (%s)' % (result['id'], display_name))


@data_commands.command('create')
@click.option('--displayName', required=True)
@click.option('--description', required=False)
@click.option('--org', required=True)
@click.option('--dataPath', required=True)
@click.option('--linked/--embedded', is_flag=True, required=True)
@click.pass_context
def _cmd_create_data_volume(ctx, displayname, description, org, datapath, linked):
    data = {}

    if org == 'me':
        org = None

    add_to_data_if_not_none(data, displayname, "display_name")
    add_to_data_if_not_none(data, org, "org")
    add_to_data_if_not_none(data, description, "description")
    add_to_data_if_not_none(data, not linked, "embedded")

    expiration = ctx.obj.config.readonly_items('data_volumes').get('expiration')
    if expiration:
        data['expiration'] = expiration

    result = ctx.obj.handle_api(ctx.obj, requests.post, "data_volumes", data)

    data_volume_id = result['id']

    datapath = __expend_and_validate_path(datapath)

    create_data_volume(data_volume_id, datapath, linked, displayname, description)

    output_result(ctx, result)


@data_commands.command('config')
@click.argument('volumeId', envvar='VOLUMEID', type=int)
@click.option('--edit', is_flag=True)
@click.pass_context
def edit_config_file(ctx, volumeid, edit):
    import subprocess

    path = os.path.join(default_data_volume_path(volumeid), 'config')

    if edit:
        subprocess.call(['edit', path])
        return

    with open(path) as f:
        click.echo(f.read())


@data_commands.command('commit')
@click.argument('volumeId', envvar='VOLUMEID', type=int)
@click.option('--message', '-m', required=False)
@click.pass_context
def commit_data_volume(ctx, volumeid, message):
    with with_repo(ctx.obj.config, volumeid) as repo:
        result = repo.commit(message)

        if 'commit_id' not in result:
            click.echo('no changeset detected', err=True)
            exit(1)

        output_result(ctx, result)


@data_commands.command('log')
@click.option('--path', '-p', required=False)
@click.pass_context
def log_data_volume(ctx, path):
    with with_repo(ctx.obj.config, path) as r:
        for entry in r.get_walker():
            commit = entry.commit
            click.echo("%s %s" % (commit.id.decode('utf8')[:7], commit.message.decode('utf8')))


def _run_add_batches(bar, repo, paths, total_files):
    def inc_bar(_):
        bar.update()

    batch_size = max(min(total_files // 100, 250), 250)  # FIXME: hardcoded

    for current_batch in get_batch_of_files_from_paths(paths, batch_size):
        porcelain.add(repo, current_batch, repo.data_volume_config.embedded, callback=inc_bar)


@data_commands.command('add')
@click.argument('volumeId', envvar='VOLUMEID', type=int)
@click.option('--files', '-f', multiple=True)
@click.option('--commit', is_flag=True, required=False)
@click.option('--message', '-m', required=False)
@click.option('--processes', default=-1, type=int, required=False)
@click.option('--no_progressbar/--enable_progressbar', default=False, is_flag=True, required=False)
@click.pass_context
def add_to_data_volume(ctx, volumeid, files, commit, message, processes, no_progressbar):
    def _validate_message_for_commit():
        if commit and not message:
            raise click.UsageError('missing --message when using --commit')

    _validate_message_for_commit()

    total_files = get_total_files_in_path(files)

    try:
        with tqdm(total=total_files, desc="Adding files", unit=' files', ncols=80, disable=no_progressbar) as bar:
            with with_repo(ctx.obj.config, volumeid) as repo:
                if processes != -1:
                    repo.data_volume_config.object_store_config['processes'] = processes

                _run_add_batches(bar, repo, files, total_files)

                if commit:
                    repo.commit(message)
    except NoCredentialsError as ex:
        click.echo('S3/boto', err=True)
        click.echo(ex.message, err=True)
        exit(1)


@data_commands.command('serve')
@click.argument('volumeId', envvar='VOLUMEID', type=int)
@click.option('--host', '-h', default='127.0.0.1', required=False)
@click.option('--port', default=3030, required=False)
@click.pass_context
def serve_data(ctx, volumeid, host, port):
    from .server import app
    click.echo('Starting server on http://{host}:{port}'.format(host=host, port=port))

    with with_repo(ctx.obj.config, volumeid) as repo:
        repo.close()
        app.config.update(dict(data_path=repo.data_path, repo_root=repo.repo_root, config_prefix=ctx.obj.config_prefix))
        app.run(host=host, port=port, debug=True)


def fill_in_vars(path, replace_vars):
    replace_vars_keys = sorted(replace_vars.keys(), reverse=True)

    for var_name in replace_vars_keys:
        var_value = replace_vars[var_name]
        path = path.replace('$' + var_name, str(var_value))
        path = path.replace('#' + var_name, str(var_value))

    return path


def safe_make_dirs(dir):
    try:
        os.makedirs(dir)
    except OSError as ex:
        if ex.errno != errno.EEXIST:
            raise


def path_elements(path):
    folders = []
    while 1:
        path, folder = os.path.split(path)

        if folder != "":
            folders.append(folder)
        else:
            if path != "":
                folders.append(path)

            break

    folders.reverse()

    return folders


def safe_rm_tree(path):
    try:
        shutil.rmtree(path)
    except OSError as ex:
        if ex.errno != errno.ENOENT:
            raise


def add_sys_var(name, value, current_vars):
    current_vars['@' + name] = value

    if name not in current_vars:
        current_vars[name] = value


def sys_vars_in_path(path, names):
    for name in names:
        if name.startswith('$'):
            if name in path:
                return True

            continue

        if '$' + name in path:
            return True

        if '$@' + name in path:
            return True

    return False


def __download_entity(config_init_dict, volume_id, dest, item):
    import google.api_core.exceptions

    config = Config(**config_init_dict)

    with with_repo(config, volume_id, read_only=True) as repo:
        try:
            _, data = repo.object_store.get_raw(item['@id'])
        except requests.exceptions.HTTPError as ex:
            if ex.response.status_code == 404:
                return

            raise
        except google.api_core.exceptions.NotFound:
            return

        phase = item.get('@phase')

        if phase is None:
            return

        _, file_extension = os.path.splitext(item['@path'])

        add_sys_var('phase', phase, item)
        add_sys_var('name', os.path.basename(item['@path']), item)
        add_sys_var('dir', os.path.dirname(item['@path']), item)
        add_sys_var('ext', os.path.basename(file_extension), item)

        item['@'] = phase

        dest_file = fill_in_vars(dest, item)

        safe_make_dirs(os.path.dirname(dest_file))

        def append_id(filename, uid):
            name, ext = os.path.splitext(filename)
            return "{name}_{uid}{ext}".format(name=name, uid=uid, ext=ext)

        if os.path.exists(dest_file):
            dest_file = append_id(dest_file, item['@id'])

        with open(dest_file, 'wb') as f:
            f.write(data)


@data_commands.command('clone')
@click.argument('volumeId', envvar='VOLUMEID', type=int)
@click.option('--dest', '-d', required=True)
@click.option('--query', '-q', required=False)
@click.option('--delete', is_flag=True, required=False)
@click.option('--batchSize', required=False, default=500)
@click.option('--processes', default=-1, type=int, required=False)
@click.option('--no_split', default=False, is_flag=True, required=False)
@click.option('--no_progressbar/--enable_progressbar', default=False, is_flag=True, required=False)
@click.pass_context
def clone_data(ctx, volumeid, dest, query, delete, batchsize, processes, no_split, no_progressbar):
    dest = __expend_and_validate_path(dest, expand_vars=False, validate_path=False)

    def find_dest_root():
        elements = path_elements(dest)

        root = []
        for element in elements:
            if element.startswith('$'):
                break

            root.append(element)

        return os.path.join(*root)

    if delete:
        root_dest = find_dest_root()
        safe_rm_tree(root_dest)

    if not no_split and not sys_vars_in_path(dest, ['phase', '$@']):
        dest = os.path.join(dest, '$@')

    if not sys_vars_in_path(dest, ['name', 'id']):
        dest = os.path.join(dest, '$@name')

    def get_next_results(start_index=0):
        current_results, current_total_data_points = repo.metadata.query(query, max_results=batchsize, start_index=start_index)

        return list(current_results), current_total_data_points

    def download_all(current_results):
        start_index = 0

        results = current_results

        while True:
            for item in results:
                item = normalize_item(item)
                config_init_dict = ctx.obj.config.init_dict

                multi_process_control.execute(
                    __download_entity, args=(config_init_dict, volumeid, dest, item), callback=lambda _: bar.update())

            if len(results) != batchsize:
                break

            start_index += len(results)

            results, _ = get_next_results()

    multi_process_control = get_multi_process_control(processes)

    try:
        with with_repo(ctx.obj.config, volumeid, read_only=True) as repo:
            begin_results, total_data_points = get_next_results()

            with tqdm(total=total_data_points, desc="Downloading", unit=' data point', ncols=80, disable=no_progressbar) as bar:
                with closing(multi_process_control):
                    try:
                        download_all(begin_results)
                    except MetadataOperationError as ex:
                        click.echo(ex.message, err=True)
                        exit(1)
                        return
    except KeyboardInterrupt:
        multi_process_control.terminate()
        multi_process_control.close()


@data_commands.group('metadata')
def metadata_commands():
    pass


def stats_from_json(now, json):
    return os.stat_result((
        0,  # mode
        0,  # inode
        0,  # device
        0,  # hard links
        0,  # owner uid
        0,  # gid
        len(json),  # size
        0,  # atime
        now,
        now,
    ))


def normalize_item(item):
    result_item = {}

    for key, val in item.items():
        if key == 'meta':
            continue

        result_item['@' + key] = val

    for key, val in item.get('meta', {}).items():
        result_item[key] = val

    return result_item


@data_commands.command('query')
@click.argument('volumeId', envvar='VOLUMEID', type=int)
@click.option('--query', '-q')
@click.option('--batchSize', required=False, default=500)
@click.pass_context
def query_metadata(ctx, volumeid, query, batchsize):
    def get_all_results():
        def get_next_results():
            current_results, current_total_data_points = repo.metadata.query(
                query, max_results=batchsize, start_index=start_index)

            return list(current_results), current_total_data_points

        total_results = []

        start_index = 0

        results, total_data_points = get_next_results()
        while True:
            total_results.extend([normalize_item(item) for item in results])

            if len(results) != batchsize:
                break

            start_index += len(results)

            results, total_data_points = get_next_results()

        return total_results

    try:
        with with_repo(ctx.obj.config, volumeid) as repo:
            all_results = get_all_results()
    except MetadataOperationError as ex:
        click.echo(ex.message, err=True)
        exit(1)
        return

    output_result(ctx, all_results)


def chunks(l, n):
    result = []
    for item in l:
        result.append(item)

        if len(result) == n:
            yield result
            result = []

    if result:
        yield result


def __json_data_from_files(data_path, files, data_per_entry):
    def rel_path_if_needed(path):
        if os.path.isabs(path):
            return os.path.relpath(path, data_path)

        return path

    for file_path in enumerate_paths(files):
        file_path = rel_path_if_needed(file_path)
        yield file_path, data_per_entry


def validate_json(ctx, param, value):
    try:
        if value is None:
            return None

        return json.loads(value)
    except ValueError:
        raise click.BadParameter('not valid json', param=param, ctx=ctx)


def __multi_line_json_from_data(data):
    def items():
        if isinstance(data, dict):
            for key, val in data.items():
                val['_sha'] = key

                yield val
        else:  # list, generators
            for key, val in data:
                val['_sha'] = key

                yield val

    result = six.BytesIO()
    for i, item in enumerate(items()):
        json_str = json.dumps(item) + '\n'
        result.write(json_str.encode())

    return result


@metadata_commands.command('add')
@click.argument('volumeId', envvar='VOLUMEID', type=int)
@click.option('--files', '-f', multiple=True)
@click.option('--data', '-d', required=False, callback=validate_json)
@click.option('--dataPoint', '-dp', multiple=True)
@click.option('--dataFile', '-df', required=False, type=click.File('rb'))
@click.option('--property', '-p', required=False, type=(str, str), multiple=True)
@click.option('--propertyInt', '-pi', required=False, type=(str, int), multiple=True)
@click.option('--propertyFloat', '-pf', required=False, type=(str, float), multiple=True)
@click.option('--update/--replace', is_flag=True, default=True, required=False)
@click.option('--no_progressbar/--enable_progressbar', default=False, is_flag=True, required=False)
@click.pass_context
def add_to_metadata(
        ctx, volumeid, files, data, datapoint, datafile, property, propertyint, propertyfloat, update, no_progressbar):

    def get_per_data_entry():
        data_per_entry = data or {}

        for props in (property, propertyint, propertyfloat):
            if not props:
                continue

            for prop_name, prop_val in props:
                data_per_entry[prop_name] = prop_val

        return data_per_entry

    def get_current_data_file():
        if datafile:
            json_data = json.load(datafile)
            return __multi_line_json_from_data(json_data)

        entries = datapoint or []
        entries.extend(files or [])

        data_list = __json_data_from_files(repo.data_path, entries, get_per_data_entry())

        return __multi_line_json_from_data(data_list)

    def upload_file_for_processing():
        data_object_name = '%s/temp/%s.json' % (volumeid, uuid4().hex)

        url = 'data_volumes/{volume_id}/gcs_urls'.format(volume_id=volumeid)

        headers = {'Content-Type': 'application/json'}

        msg = {
            'methods': 'PUT',
            'paths': [data_object_name],
            'content_type': 'application/json'
        }

        result = ctx.obj.handle_api(ctx.obj, requests.post, url, msg)

        put_url = result['put'][0]

        file_obj = get_current_data_file()

        with tqdm(total=file_obj.tell(), desc="Uploading", unit=' KB', ncols=80, disable=no_progressbar, unit_scale=True) as bar:
            do_upload(gcs_auth, None, data_object_name, file_obj, headers, head_url=None, put_url=put_url, callback=lambda c: bar.update(c))

        return data_object_name

    with with_repo(ctx.obj.config, volumeid) as repo:
        gcs_auth = repo.data_volume_config.object_store_config.get('auth')

        validate_repo_data_path(repo, volumeid)

        metadata = repo.metadata

        object_name = upload_file_for_processing()

        start_time = datetime.datetime.utcnow()
        click.echo('Processing Metadata')
        metadata.add_data(data='gs://' + object_name)
        click.echo('Processing Done (%s)' % (datetime.datetime.utcnow() - start_time))



# -*- coding: utf8 -*-
import os

import fudge
from click.testing import CliRunner
from fudge.inspector import arg

from mali import cli
from mali_commands.utilities import source_tracking
from .base import BaseCliTest


class TestAuth(BaseCliTest):
    # TODO: add a test with git init
    def test_code_no_source_not_even_in_dir(self):
        user_id, org_id, org, repo = (self.some_random_shit(size=10) for _ in range(4))
        hurl = 'https://simhub.com/{}/{}'.format(org, repo)
        gurl = 'git@simhub.com:{}/{}.git'.format(org, repo)
        runner = CliRunner()
        with runner.isolated_filesystem():
            result = runner.invoke(
                cli,
                ["code", "track", "--org", org_id, "--tracking-repository", hurl + '1'],
                catch_exceptions=True)
            self.assertEquals(
                result.exception.args[0],
                "Couldn't determinate the remote origin of the current directory, please provide --source-repository url")

    def test_code_bad_source(self):
        user_id, org_id, org, repo = (self.some_random_shit(size=10) for _ in range(4))
        hurl = 'https://simhub.com/{}/{}'.format(org, repo)
        gurl = 'git@simhub.com:{}/{}.git'.format(org, repo)
        s_url = self.some_random_shit(size=10)

        runner = CliRunner()
        result = runner.invoke(
            cli,
            ["code", "track", "--org", org_id, "--source-repository", s_url, "--tracking-repository", hurl + '1'],
            catch_exceptions=True)
        self.assertEquals(
            result.exception.args[0],
            "Source repository: {} is invalid".format(s_url))

    def test_code_bad_target(self):
        user_id, org_id, org, repo = (self.some_random_shit(size=10) for _ in range(4))
        hurl = 'https://simhub.com/{}/{}'.format(org, repo)
        gurl = 'git@simhub.com:{}/{}.git'.format(org, repo)
        s_url = self.some_random_shit(size=10)

        runner = CliRunner()
        result = runner.invoke(
            cli,
            ["code", "track", "--org", org_id, "--source-repository", gurl, "--tracking-repository", s_url],
            catch_exceptions=True)
        self.assertEquals(
            result.exception.args[0],
            "Tracking repository: {} is invalid".format(s_url))

    def test_code_same_url(self):
        user_id, org_id, org, repo = (self.some_random_shit(size=10) for _ in range(4))
        hurl = 'https://simhub.com/{}/{}'.format(org, repo)
        gurl = 'git@simhub.com:{}/{}.git'.format(org, repo)

        runner = CliRunner()
        with runner.isolated_filesystem():
            result = runner.invoke(
                cli,
                ["code", "track", "--org", org_id, "--source-repository", gurl, "--tracking-repository", gurl],
                catch_exceptions=True)
            self.assertEquals(
                result.exception.args[0],
                "The repository repository can't be the same as the source one")
        with runner.isolated_filesystem():
            result = runner.invoke(
                cli,
                ["code", "track", "--org", org_id, "--source-repository", gurl, "--tracking-repository", hurl],
                catch_exceptions=True)
            self.assertEquals(
                result.exception.args[0],
                "The repository repository can't be the same as the source one")
        with runner.isolated_filesystem():
            result = runner.invoke(
                cli,
                ["code", "track", "--org", org_id, "--source-repository", hurl, "--tracking-repository", hurl],
                catch_exceptions=True)
            self.assertEquals(
                result.exception.args[0],
                "The repository repository can't be the same as the source one")
        with runner.isolated_filesystem():
            result = runner.invoke(
                cli,
                ["code", "track", "--org", org_id, "--source-repository", hurl, "--tracking-repository", gurl],
                catch_exceptions=True)
            self.assertEquals(
                result.exception.args[0],
                "The repository repository can't be the same as the source one")

    def test_code_diff_ogs(self):
        user_id, org_id, org, repo = (self.some_random_shit(size=10) for _ in range(4))
        hurl = 'https://simhub.com/{}/{}'.format(org, repo)
        gurl = 'git@simhub.com:{}{}/{}xx.git'.format(org, self.some_random_shit(size=2), repo)

        runner = CliRunner()
        result = runner.invoke(
            cli,
            ["code", "track", "--org", org_id, "--source-repository", hurl, "--tracking-repository", gurl],
            catch_exceptions=True)
        self.assertEquals(
            result.exception.args[0],
            "The tracking repository must belong to the same organization as the source (/{})".format(org))

        hurl = 'https://simhub.com/{}{}/{}x'.format(org, self.some_random_shit(size=2), repo)
        gurl = 'git@simhub.com:{}/{}.git'.format(org, repo)

        result = runner.invoke(
            cli,
            ["code", "track", "--org", org_id, "--source-repository", gurl, "--tracking-repository", hurl],
            catch_exceptions=True)
        self.assertEquals(
            result.exception.args[0],
            "The tracking repository must belong to the same organization as the source (/{})".format(org))

    @fudge.patch('mali_commands.commons.handle_api')
    def test_code_works_with_explicit_source(self, handle_api):
        user_id, org_id, org, repo = (self.some_random_shit(size=10) for _ in range(4))
        hurl = 'https://simhub.com/{}/{}'.format(org, repo)
        gurl = 'git@simhub.com:{}/{}_1.git'.format(org, repo)
        encoded_url = source_tracking.remote_to_template(hurl)

        handle_api.expects_call().with_matching_args(
            arg.any(),  # ctx
            arg.any(),  # http method
            '{}/code/track?src={}'.format(org_id, encoded_url),
            {
                'target': gurl
            }
        ).returns({})

        runner = CliRunner()
        result = runner.invoke(
            cli,
            ["code", "track", "--org", org_id, "--source-repository", hurl, "--tracking-repository", gurl],
            catch_exceptions=False)
        self.assertEquals(result.exit_code, 0)

    @fudge.patch('mali_commands.commons.handle_api')
    def test_code_works_without_explicit_source(self, handle_api):
        user_id, org_id, org, repo = (self.some_random_shit(size=10) for _ in range(4))
        hurl = 'https://simhub.com/{}/{}'.format(org, repo)
        gurl = 'git@simhub.com:{}/{}_1.git'.format(org, repo)
        encoded_url = source_tracking.remote_to_template(hurl)

        runner = CliRunner()
        with runner.isolated_filesystem():
            import git
            g = git.Repo.init('.')
            g.git.remote('add', 'origin', hurl)
            handle_api.expects_call().with_matching_args(
                arg.any(),  # ctx
                arg.any(),  # http method
                '{}/code/track?src={}'.format(org_id, encoded_url),
                {
                    'target': gurl
                }
            ).returns({})

            result = runner.invoke(
                cli,
                ["code", "track", "--org", org_id, "--tracking-repository", gurl],
                catch_exceptions=False)
            self.assertEquals(result.exit_code, 0)

    def testget_org_from_remote_url(self):
        from mali_commands.utilities import source_tracking
        x = source_tracking.get_org_from_remote_url(os.getcwd())
        self.assertIsNone(x)
        repo = '://simhub.com/{}/{}'.format(self.some_random_shit(size=6), self.some_random_shit(size=9))
        x = source_tracking.remote_to_template('http' + repo)
        self.assertEquals(x, 'https' + repo)



# -*- coding: utf8 -*-
import logging
from collections import OrderedDict

import botocore

from ...gcs_utils import gcs_credentials, do_upload, do_delete_all, do_download, s3_moniker
from ...connection_mixin import ConnectionMixin
from ...dulwich.object_store import BaseObjectStore
from ...dulwich.objects import hex_to_filename, sha_to_hex, Blob

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


def return_request_result(response, content):
    return response, content


class GCSObjectStore(ConnectionMixin, BaseObjectStore):
    def __init__(self, connection, use_multiprocess=True, processes=-1):
        super(GCSObjectStore, self).__init__(connection)
        self.__upload_pool = None
        self._use_multiprocess = use_multiprocess
        self._multi_process_control = None
        self._processes = processes
        self._object_store_auth = connection.data_volume_config.object_store_config.get('auth')
        self.__bucket_name = connection.data_volume_config.object_store_config.get('bucket_name')
        self.__volume_id = self._connection.data_volume_config.volume_id
        self._signed_url_service = None

    def delete_all(self, max_files=None):
        return do_delete_all(self.__bucket_name, self.__volume_id, max_files)

    @classmethod
    def get_content_headers(cls, content_type=None):

        headers = OrderedDict()
        if content_type:
            headers['Content-Type'] = content_type

        headers['x-goog-acl'] = 'public-read'
        headers['x-goog-if-generation-match'] = '0'

        return headers

    @property
    def processes(self):
        return self._processes if self.is_multiprocess else 1

    @processes.setter
    def processes(self, value):
        self._processes = value

    @property
    def is_multiprocess(self):
        return self._use_multiprocess and self._processes != 1

    def close(self):
        logging.debug('%s closing', self.__class__)
        if self._multi_process_control is not None:
            self._multi_process_control.close()

        logging.debug('%s closed', self.__class__)

    @classmethod
    def _get_shafile_path(cls, sha):
        # Check from object dir
        return hex_to_filename('objects', sha)

    @classmethod
    def on_upload_error(cls, ex):
        raise ex

    def __init_multi_process_if_needed(self):
        if self._multi_process_control is None:
            from ...multi_process_control import get_multi_process_control

            self._multi_process_control = get_multi_process_control(self.processes)

    def upload(self, obj, content_type=None, head_url=None, put_url=None, callback=None):
        from mali_commands.utilities import get_content_type

        object_name = self._get_shafile_path(obj.blob.id)

        if self.__bucket_name and not self.__bucket_name.startswith(s3_moniker):
            credentials = gcs_credentials()
        else:
            credentials = None

        content_type = content_type or get_content_type(obj.full_path)
        headers = GCSObjectStore.get_content_headers(content_type)

        object_name = '%s/%s' % (self.__volume_id, object_name)

        args = (
            self._object_store_auth,
            self.__bucket_name, object_name, obj.full_path, headers, head_url, put_url, credentials)

        self.__init_multi_process_if_needed()

        self._multi_process_control.execute(do_upload, args=args, callback=callback)

    def add_object(self, obj):

        """Add a single object to this object store.

        :param obj: Object to add
        """
        self.upload(obj)

    def _get_loose_object(self, sha):
        logger.debug('get object %s', sha)
        object_name = '%s/%s' % (self.__volume_id, self._get_shafile_path(sha))

        data = do_download(
            self._object_store_auth, self.__bucket_name, object_name, signed_url_service=self._signed_url_service)

        blob = Blob()
        blob.set_raw_chunks([data], sha)
        return blob

    def get_raw(self, name):
        """Obtain the raw text for an object.

        :param name: sha for the object.
        :return: tuple with numeric type and object contents.
        """
        hex_sha = name

        if len(name) != 40 and len(name) != 20:
            raise AssertionError("Invalid object name %r" % name)

        if hex_sha is None:
            hex_sha = sha_to_hex(name)

        ret = self._get_loose_object(hex_sha)
        if ret is not None:
            return ret.type_num, ret.as_raw_string()

        raise KeyError(hex_sha)

    @property
    def packs(self):
        raise NotImplementedError(self.packs)

    def __iter__(self):
        raise NotImplementedError(self.__iter__)

    def add_objects(self, objects, callback=None):
        for obj in objects:
            self.upload(obj, callback=callback)

    def contains_packed(self, sha):
        raise NotImplementedError(self.contains_packed)

    def contains_loose(self, sha):
        raise NotImplementedError(self.contains_loose)



import logging

from six.moves.urllib import parse


class GitError(Exception):
    pass


class GitExpando(object):
    pass


def ssh_to_https(url):
    if '@' in url:
        return 'https://' + (url.split('@')[1].strip()).replace(':', '/')
    return url


def enforce_https(url):
    if url.startswith('http://'):
        return 'https://' + url[len('http://'):]
    return url


def remove_dot_git(url):
    return url if not url.lower().endswith('.git') else url[:-4]


def remote_to_template(remote_url):
    return remove_dot_git(enforce_https(ssh_to_https(remote_url.lower())))


def get_remote_url():
    try:
        import git
        repo = git.Repo('.', search_parent_directories=True)
        if len(repo.remotes) > 0:
            remote = list(repo.remotes)[0]
            return remote_to_template(remote.url)
    except Exception as gr:
        logging.error("Failed to get remote git repository of current folder. {}".format(str(gr)))
    return None


def remote_url_obj(url):
    if url is None:
        return None
    obj = parse.urlparse(remote_to_template(url))
    if obj.hostname is None or obj.hostname == '':
        return None
    return obj


def get_org_from_remote_url(url):
    url_obj = remote_url_obj(url)
    if url_obj is None:
        return None
    return '/'.join(url_obj.path.split('/')[:-1]).lower()


def validate_diff_urls(src, trg):
    src = remote_to_template(src)
    trg = remote_to_template(trg)
    return src == trg


def validate_remote_urls_org(src, trg):
    src_org = get_org_from_remote_url(src)
    remote_org = get_org_from_remote_url(trg)

    return src_org == remote_org



# -*- coding: utf8 -*-
import json
import logging

import copy

from ..backend_mixin import BackendMixin
from .base_metadata_db import BaseMetadataDB, MetadataOperationError
from six.moves.urllib.parse import urlencode


class BackendMetadataDB(BackendMixin, BaseMetadataDB):
    def __init__(self, connection, config, handle_api):
        super(BackendMetadataDB, self).__init__(connection, config, handle_api)

    def _create_table(self):
        pass

    def _query_head_data(self, sha_list):
        with self._connection.get_cursor() as session:
            url = 'data_volumes/%s/metadata/head' % self._volume_id
            msg = {
                'sha': sha_list,
            }

            result = self._handle_api(self._config, session.post, url, msg)

            for data_item in result.get('metadata_json') or []:
                yield json.loads(data_item)

    def _add_missing_columns(self, data_object):
        pass

    def get_data_for_commit(self, sha, commit_sha):
        raise NotImplementedError(self.get_data_for_commit)

    def _add_data(self, data):
        pass

    def add_data(self, data):
        if not data:
            logging.debug('no data provided')
            return

        logging.debug('add data %s', data)

        with self._connection.get_cursor() as session:
            url = 'data_volumes/%s/metadata/head/add' % self._volume_id

            msg = {
                'metadata_url': data,
            }

            return self._handle_api(self._config, session.post, url, msg)

    def query(self, query_text, max_results, start_index):
        def _iter_data():
            for data_point in result.get('data_points') or []:
                result_data_point = {}

                for key, val in data_point.items():
                    if key == 'meta':
                        meta = result_data_point.setdefault('meta', {})
                        for meta_key_val in data_point['meta']:
                            meta[meta_key_val['key']] = meta_key_val.get('val')
                    else:
                        result_data_point[key] = val

                yield result_data_point

        with self._connection.get_cursor() as session:
            version_query = query_text if query_text else '@version:head'

            params = {
                'query': version_query
            }

            if max_results:
                params['max_results'] = max_results

            if start_index:
                params['start_index'] = start_index

            url = 'data_volumes/%s/query/?%s' % (self._volume_id, urlencode(params))

            result = self._handle_api(self._config, session.get, url)

            if not result['ok']:
                raise MetadataOperationError(result['error'])

            return _iter_data(), int(result.get('total_data_points', 0))

    def _query(self, sql_vars, select_fields, where, max_results, start_index):
        raise NotImplementedError(self._query)

    def get_all_data(self, sha):
        raise NotImplementedError(self.get_all_data)

    def end_commit(self):
        raise NotImplementedError(self.end_commit)

    def begin_commit(self, commit_sha, tree_id):
        raise NotImplementedError(self.begin_commit)



# -*- coding: utf8 -*-
import os
import errno
import shutil


def create_dir(dirname):
    if not dirname or os.path.exists(dirname):
        return

    try:
        os.makedirs(dirname)
    except OSError as exc:
        if exc.errno != errno.EEXIST:
            raise


def remove_dir(dirname):
    try:
        shutil.rmtree(dirname)
    except (OSError, IOError):
        pass


def flatten_dir(root_dir):
    root_walk = os.walk(root_dir)
    _, top_level_dirs, _ = next(root_walk)

    for dirpath, _, filenames in root_walk:
        for filename in filenames:
            filepath = os.path.join(dirpath, filename)
            shutil.move(filepath, root_dir)

    for top_level_dir in top_level_dirs:
        remove_dir(os.path.join(root_dir, top_level_dir))
